import math
import torch.nn.functional as F  # noqa: N812
from torch import nn
import torch


class DConv(nn.Module):
    def __init__(self, dim):
        super().__init__()
        self.conv = nn.Conv2d(dim, dim, kernel_size=7, padding=7 // 2, groups=dim)

    def forward(self, x, res):
        B, N, C = x.shape
        H, W = res
        x = x.transpose(1, 2).view(B, C, H, W)
        return self.conv(x).flatten(2).transpose(1, 2)


class FLPVT2(nn.Module):
    def __init__(self, dim, num_heads=8, qkv_bias=True, focusing_factor=3, kernel_size=5):
        super().__init__()
        assert dim % num_heads == 0, f'dim {dim} should be divided by num_heads {num_heads}.'

        self.dim = dim
        self.num_heads = num_heads
        head_dim = dim // num_heads
        self.q = nn.Linear(dim, dim, bias=qkv_bias)
        self.kv = nn.Linear(dim, dim * 2, bias=qkv_bias)
        self.proj = nn.Linear(dim, dim)

        self.focusing_factor = nn.Parameter(torch.ones(dim) * focusing_factor)
        self.dwc = nn.Conv2d(in_channels=head_dim, out_channels=head_dim, kernel_size=kernel_size, groups=head_dim, padding=kernel_size // 2)
        self.scale = nn.Parameter(torch.zeros(dim))
        self.apply(self._init_weights)

    def _init_weights(self, m):
        if isinstance(m, nn.Linear):
            torch.nn.init.trunc_normal_(m.weight, std=0.02)
            if isinstance(m, nn.Linear) and m.bias is not None:
                nn.init.constant_(m.bias, 0)
        elif isinstance(m, nn.RMSNorm):
            # nn.init.constant_(m.bias, 0)
            nn.init.constant_(m.weight, 1.0)
        elif isinstance(m, nn.LayerNorm):
            nn.init.constant_(m.bias, 0)
            nn.init.constant_(m.weight, 1.0)
        elif isinstance(m, nn.Conv2d):
            fan_out = m.kernel_size[0] * m.kernel_size[1] * m.out_channels
            fan_out //= m.groups
            m.weight.data.normal_(0, math.sqrt(2.0 / fan_out))
            if m.bias is not None:
                m.bias.data.zero_()

    def forward(self, x, resol):
        H, W = resol
        B, N, C = x.shape
        q = self.q(x)

        kv = self.kv(x).reshape(B, -1, 2, C).permute(2, 0, 1, 3)
        k, v = kv[0], kv[1]
        n = k.shape[1]
        scale = torch.functional.F.softplus(self.scale)
        q = torch.functional.F.relu(q) + 1e-6
        k = torch.functional.F.relu(k) + 1e-6
        q = q / scale
        k = k / scale
        q_norm = q.norm(dim=-1, keepdim=True)
        k_norm = k.norm(dim=-1, keepdim=True)
        q = q**self.focusing_factor
        k = k**self.focusing_factor
        q = (q / q.norm(dim=-1, keepdim=True)) * q_norm
        k = (k / k.norm(dim=-1, keepdim=True)) * k_norm

        q = q.reshape(B, N, self.num_heads, -1).permute(0, 2, 1, 3)
        k = k.reshape(B, n, self.num_heads, -1).permute(0, 2, 1, 3)
        v = v.reshape(B, n, self.num_heads, -1).permute(0, 2, 1, 3)

        z = 1 / (q @ k.mean(dim=-2, keepdim=True).transpose(-2, -1) + 1e-6)
        kv = (k.transpose(-2, -1) * (n**-0.5)) @ (v * (n**-0.5))
        x = q @ kv * z
        x = x.transpose(1, 2).reshape(B, N, C)
        v = v.reshape(B * self.num_heads, H, W, -1).permute(0, 3, 1, 2)
        x = x + self.dwc(v).reshape(B, C, N).permute(0, 2, 1)

        x = self.proj(x)

        return x


class GatedCNNBlock(nn.Module):
    r"""Our implementation of Gated CNN Block: https://arxiv.org/pdf/1612.08083
    Args:
        conv_ratio: control the number of channels to conduct depthwise convolution.
            Conduct convolution on partial channels can improve paraitcal efficiency.
            The idea of partial channels is from ShuffleNet V2 (https://arxiv.org/abs/1807.11164) and
            also used by InceptionNeXt (https://arxiv.org/abs/2303.16900) and FasterNet (https://arxiv.org/abs/2303.03667)
    """

    def __init__(self, dim, expansion_ratio=8 / 3, conv_ratio=1.0, drop_path=0.0, att=False):
        super().__init__()
        if att:
            expansion_ratio = 1.5
        self.norm = nn.RMSNorm(dim, eps=1e-6)
        hidden = int(expansion_ratio * dim)
        self.fc1 = nn.Linear(dim, hidden * 2)
        self.act = nn.Mish()
        conv_channels = int(conv_ratio * dim)
        self.split_indices = (hidden, hidden - conv_channels, conv_channels)
        self.conv = FLPVT2(conv_channels) if att else DConv(conv_channels)
        self.fc2 = nn.Linear(hidden, dim)
        self.apply(self._init_weights)

    @staticmethod
    def _init_weights(m):
        if isinstance(m, nn.Conv2d | nn.Linear):
            nn.init.trunc_normal_(m.weight, std=0.02)
            if m.bias is not None:
                nn.init.constant_(m.bias, 0)

    def forward(self, x, res):
        # shortcut = x # [B, H, W, C]
        x = self.norm(x)
        g, i, c = torch.split(self.fc1(x), self.split_indices, dim=-1)
        c = self.conv(c, res)
        x = self.fc2(self.act(g) * torch.cat((i, c), dim=-1))
        return x  # + shortcut


class Blocks(nn.Module):
    def __init__(self, dim, n_block, att=False):
        super().__init__()
        self.gated = nn.ModuleList([GatedCNNBlock(dim, att=att) for _ in range(n_block)])

    def forward(self, x):
        B, C, H, W = x.shape
        x = x.flatten(2).transpose(1, 2)
        for block in self.gated:
            x = block(x, (H, W)) + x
        return x.transpose(1, 2).view(B, C, H, W)


class Downsample(nn.Module):
    def __init__(self, n_feat):
        super(Downsample, self).__init__()

        self.body = nn.Sequential(nn.Conv2d(n_feat, n_feat // 2, kernel_size=3, stride=1, padding=1, bias=True), nn.PixelUnshuffle(2))

    def forward(self, x):
        return self.body(x)


class Upsample(nn.Module):
    def __init__(self, n_feat):
        super(Upsample, self).__init__()

        self.body = nn.Sequential(nn.Conv2d(n_feat, n_feat * 2, kernel_size=3, stride=1, padding=1, bias=True), nn.PixelShuffle(2))

    def forward(self, x):
        return self.body(x)


class GateR(nn.Module):
    def __init__(self, dim=48, in_ch=3, num_blocks=(3, 6, 6, 10, 6, 6, 3), latent_att=False):
        super().__init__()
        self.in_to_dim = nn.Conv2d(in_ch, dim, 3, 1, 1)

        self.enc0 = Blocks(dim, num_blocks[0])
        self.enc1 = nn.Sequential(Downsample(dim), Blocks(dim * 2, num_blocks[1]))
        self.enc2 = nn.Sequential(Downsample(dim * 2), Blocks(dim * 4, num_blocks[2]))

        self.latent = nn.Sequential(Downsample(dim * 4), Blocks(dim * 8, num_blocks[3], latent_att), Upsample(dim * 8))

        self.dec0 = nn.Sequential(nn.Conv2d(dim * 8, dim * 4, 1), Blocks(dim * 4, num_blocks[4]), Upsample(dim * 4))
        self.dec1 = nn.Sequential(nn.Conv2d(dim * 4, dim * 2, 1), Blocks(dim * 2, num_blocks[5]), Upsample(dim * 2))
        self.dec2 = nn.Sequential(Blocks(dim * 2, num_blocks[6]))
        self.dim_to_ch = nn.Sequential(nn.Conv2d(dim * 2, dim, 3, 1, 1), nn.Conv2d(dim, in_ch, 3, 1, 1))

        # self.dec0 =

    def check_img_size(self, x, resolution: tuple[int, int]):
        h, w = resolution
        scaled_size = 8
        mod_pad_h = (scaled_size - h % scaled_size) % scaled_size
        mod_pad_w = (scaled_size - w % scaled_size) % scaled_size
        return F.pad(x, (0, mod_pad_w, 0, mod_pad_h), 'reflect')

    def forward(self, x):
        B, C, H, W = x.shape
        x = self.check_img_size(x, (H, W))
        enc = self.in_to_dim(x)
        enc0 = self.enc0(enc)
        enc1 = self.enc1(enc0)
        enc2 = self.enc2(enc1)

        latent = self.latent(enc2)

        dec0 = self.dec0(torch.cat([latent, enc2], dim=1))
        dec1 = self.dec1(torch.cat([dec0, enc1], dim=1))
        dec2 = self.dec2(torch.cat([dec1, enc0], dim=1))
        return (self.dim_to_ch(dec2) + x)[:, :, :H, :W]



import re
from typing import Mapping

from .arch import eimn
from ...factory import KeyCondition, Architecture
from ...utilities.state_dict import get_seq_len, pixelshuffle_scale


class eimnArch(Architecture[eimn]):
    def __init__(self):
        super().__init__(
            uid='eimn',
            detect=KeyCondition.has_all(
                'head.0.weight',
                'head.0.bias',
                'tail.0.weight',
                'tail.0.bias',
                'block1.0.layer_scale_1',
                'block1.0.layer_scale_2',
                'block1.0.norm1.weight',
                'block1.0.norm1.bias',
                'block1.0.norm1.running_mean',
                'block1.0.norm1.running_var',
                'block1.0.norm1.num_batches_tracked',
                'block1.0.attn.region.weight',
                'block1.0.attn.region.bias',
                'block1.0.attn.spatial_1.weight',
                'block1.0.attn.spatial_1.bias',
                'block1.0.attn.spatial_2.weight',
                'block1.0.attn.spatial_2.bias',
                'block1.0.attn.fusion.weight',
                'block1.0.attn.fusion.bias',
                'block1.0.attn.proj_value.0.weight',
                'block1.0.attn.proj_value.0.bias',
                'block1.0.attn.proj_query.0.weight',
                'block1.0.attn.proj_query.0.bias',
                'block1.0.attn.out.weight',
                'block1.0.attn.out.bias',
                'block1.0.norm2.weight',
                'block1.0.norm2.bias',
                'block1.0.norm2.running_mean',
                'block1.0.norm2.running_var',
                'block1.0.norm2.num_batches_tracked',
                'block1.0.mlp.linear_in.weight',
                'block1.0.mlp.linear_in.bias',
                'block1.0.mlp.SAL.weight',
                'block1.0.mlp.SAL.bias',
                'block1.0.mlp.linear_out.weight',
                'block1.0.mlp.linear_out.bias',
                'block1.0.mlp.DFFM.norm.weight',
                'block1.0.mlp.DFFM.norm.bias',
                'block1.0.mlp.DFFM.global_reduce.weight',
                'block1.0.mlp.DFFM.global_reduce.bias',
                'block1.0.mlp.DFFM.local_reduce.weight',
                'block1.0.mlp.DFFM.local_reduce.bias',
                'block1.0.mlp.DFFM.channel_expand.weight',
                'block1.0.mlp.DFFM.channel_expand.bias',
                'block1.0.mlp.DFFM.spatial_expand.weight',
                'block1.0.mlp.DFFM.spatial_expand.bias',
                'norm1.weight',
                'norm1.bias',
            ),
        )

    def load(self, state: Mapping[str, object]):
        pattern = r'block(\d+)'
        numbers = [int(re.search(pattern, s).group(1)) for s in state.keys() if re.search(pattern, s)]
        num_stages = max(numbers)
        depths = get_seq_len(state, 'block1')
        mlp_ratio_shape = state['block1.0.mlp.linear_in.weight'].shape
        mlp_ratio = mlp_ratio_shape[0] // 2 / mlp_ratio_shape[1]
        embed_dim = state['head.0.weight'].shape[0]
        scale = pixelshuffle_scale(state['tail.0.weight'].shape[0], 3)

        model = eimn(
            embed_dims=embed_dim,
            scale=scale,
            depths=depths,
            mlp_ratios=mlp_ratio,
            num_stages=num_stages,
        )

        return self._enhance_model(model=model, in_channels=3, out_channels=3, upscale=scale, name='EIMN')



from typing import Mapping

from .arch import SRVGGNetCompact
from ...factory import Architecture, KeyCondition
from ...utilities.state_dict import get_seq_len, pixelshuffle_scale


class CompactArch(Architecture[SRVGGNetCompact]):
    def __init__(self):
        super().__init__(
            uid='Compact',
            detect=KeyCondition.has_all(
                'body.0.weight',
                'body.1.weight',
            ),
        )

    def load(self, state_dict: Mapping[str, object]):
        state = state_dict

        highest_num = get_seq_len(state, 'body') - 1

        in_nc = state['body.0.weight'].shape[1]
        num_feat = state['body.0.weight'].shape[0]
        num_conv = (highest_num - 2) // 2

        pixelshuffle_shape = state[f'body.{highest_num}.bias'].shape[0]

        scale = pixelshuffle_scale(pixelshuffle_shape, in_nc)
        model = SRVGGNetCompact(
            num_in_ch=in_nc,
            num_out_ch=in_nc,
            num_feat=num_feat,
            num_conv=num_conv,
            upscale=scale,
        )

        return self._enhance_model(model=model, in_channels=in_nc, out_channels=in_nc, upscale=scale, name='Compact')



import math
from typing import Mapping

from .arch import mosr as MoSR
from ...factory import KeyCondition, Architecture
from ...utilities.state_dict import get_seq_len, pixelshuffle_scale, dysample_scale


class MoSRArch(Architecture[MoSR]):
    def __init__(self):
        super().__init__(
            uid='MoSR',
            detect=KeyCondition.has_all(
                'gblocks.0.weight',
                'gblocks.0.bias',
                'gblocks.1.norm.weight',
                'gblocks.1.norm.bias',
                'gblocks.1.fc1.weight',
                'gblocks.1.fc1.bias',
                'gblocks.1.conv.weight',
                'gblocks.1.conv.bias',
                'gblocks.1.fc2.weight',
                'gblocks.1.fc2.bias',
            ),
        )

    def load(self, state: Mapping[str, object]):
        # Get values from state
        n_block = get_seq_len(state, 'gblocks') - 6
        in_ch = state['gblocks.0.weight'].shape[1]
        dim = state['gblocks.0.weight'].shape[0]

        # Calculate expansion ratio and convolution ratio
        expansion_ratio = (state['gblocks.1.fc1.weight'].shape[0] / state['gblocks.1.fc1.weight'].shape[1]) / 2
        conv_ratio = state['gblocks.1.conv.weight'].shape[0] / dim
        kernel_size = state['gblocks.1.conv.weight'].shape[2]
        # Determine upsampler type and calculate upscale
        if 'upsampler.init_pos' in state:
            upsampler = 'dys'
            out_ch = state['upsampler.end_conv.weight'].shape[0]
            upscale = dysample_scale(state['upsampler.offset.weight'].shape[0])
        elif 'upsampler.in_to_k.weight' in state:
            upsampler = 'gps'
            out_ch = in_ch
            upscale = math.isqrt(state['upsampler.in_to_k.weight'].shape[0] // 8 // out_ch)
        else:
            upsampler = 'ps'
            out_ch = in_ch
            upscale = pixelshuffle_scale(state['upsampler.0.weight'].shape[0], out_ch)

        model = MoSR(
            in_ch=in_ch,
            out_ch=out_ch,
            n_block=n_block,
            upscale=upscale,
            dim=dim,
            upsampler=upsampler,
            expansion_ratio=expansion_ratio,
            conv_ratio=conv_ratio,
            kernel_size=kernel_size,
        )

        return self._enhance_model(model=model, in_channels=in_ch, out_channels=out_ch, upscale=upscale, name='MoSR')



from typing import Mapping, Union

from .plksr import plksr
from .rplksr import realplksr
from ...factory import Architecture, KeyCondition
from ...utilities.state_dict import get_seq_len, pixelshuffle_scale

_PLKSR = Union[plksr, realplksr]


class PLKSRArch(Architecture[_PLKSR]):
    def __init__(self):
        super().__init__(
            uid='PLKSR',
            detect=KeyCondition.has_all(
                'feats.0.weight',
                KeyCondition.has_any(
                    'feats.1.lk.conv.weight',
                    'feats.1.lk.convs.0.weight',
                    'feats.1.lk.mn_conv.weight',
                ),
                'feats.1.refine.weight',
                KeyCondition.has_any(
                    'feats.1.channe_mixer.0.weight',
                    'feats.1.channel_mixer.0.weight',
                ),
            ),
        )

    def load(self, state_dict: Mapping[str, object]):
        kernel_size = 17
        in_nc = state_dict['feats.0.weight'].shape[1]
        out_nc = in_nc

        dim = state_dict['feats.0.weight'].shape[0]
        name = 'PLKSR'
        total_feat_layers = get_seq_len(state_dict, 'feats')
        use_ea = 'feats.1.attn.f.0.weight' in state_dict
        scale = pixelshuffle_scale(state_dict[f'feats.{total_feat_layers - 1}.weight'].shape[0], out_nc)
        if 'feats.1.channe_mixer.0.weight' in state_dict:
            # Yes, the normal version has this typo.
            n_blocks = total_feat_layers - 2

            # ccm_type
            mixer_0_shape = state_dict['feats.1.channe_mixer.0.weight'].shape[2]
            mixer_2_shape = state_dict['feats.1.channe_mixer.2.weight'].shape[2]
            if mixer_0_shape == 3 and mixer_2_shape == 1:
                ccm_type = 'CCM'
            elif mixer_0_shape == 3 and mixer_2_shape == 3:
                ccm_type = 'DCCM'
            elif mixer_0_shape == 1 and mixer_2_shape == 3:
                ccm_type = 'ICCM'
            else:
                raise ValueError('Unknown CCM type')

            # lk_type
            lk_type = 'PLK'
            use_max_kernel: bool = False
            sparse_kernels = [5, 5, 5, 5]
            sparse_dilations = [1, 2, 3, 4]
            with_idt: bool = False  # undetectable

            if 'feats.1.lk.conv.weight' in state_dict:
                # PLKConv2d
                lk_type = 'PLK'
                kernel_size = state_dict['feats.1.lk.conv.weight'].shape[2]
                split_ratio = state_dict['feats.1.lk.conv.weight'].shape[0] / dim
            elif 'feats.1.lk.convs.0.weight' in state_dict:
                # SparsePLKConv2d
                lk_type = 'SparsePLK'
                split_ratio = state_dict['feats.1.lk.convs.0.weight'].shape[0] / dim
                # Detecting other parameters for SparsePLKConv2d is praticaly impossible.
                # We cannot detect the values of sparse_dilations at all, we only know it has the same length as sparse_kernels.
                # Detecting the values of sparse_kernels is possible, but we don't know its length exactly, because it's `len(sparse_kernels) = len(convs) - (1 if use_max_kernel else 0)`.
                # However, we cannot detect use_max_kernel, because the convolutions it adds when enabled look the same as the other convolutions.
                # So I give up.
            elif 'feats.1.lk.mn_conv.weight' in state_dict:
                # RectSparsePLKConv2d
                lk_type = 'RectSparsePLK'
                kernel_size = state_dict['feats.1.lk.mn_conv.weight'].shape[2]
                split_ratio = state_dict['feats.1.lk.mn_conv.weight'].shape[0] / dim
            else:
                raise ValueError('Unknown LK type')

            model = plksr(
                dim=dim,
                n_blocks=n_blocks,
                upscaling_factor=scale,
                ccm_type=ccm_type,
                kernel_size=kernel_size,
                split_ratio=split_ratio,
                lk_type=lk_type,
                use_max_kernel=use_max_kernel,
                sparse_kernels=sparse_kernels,
                sparse_dilations=sparse_dilations,
                with_idt=with_idt,
                use_ea=use_ea,
            )
        elif 'feats.1.channel_mixer.0.weight' in state_dict:
            # and RealPLKSR doesn't. This makes it really convenient to detect.
            name = 'RealPLKSR'

            n_blocks = total_feat_layers - 3
            kernel_size = state_dict['feats.1.lk.conv.weight'].shape[2]
            split_ratio = state_dict['feats.1.lk.conv.weight'].shape[0] / dim
            dysample = False
            if 'to_img.init_pos' in state_dict:
                dysample = True
            model = realplksr(
                dim=dim,
                upscaling_factor=scale,
                n_blocks=n_blocks,
                kernel_size=kernel_size,
                split_ratio=split_ratio,
                use_ea=use_ea,
                norm_groups=4,  # un-detectable
                dysample=dysample,
            )
        else:
            raise ValueError('Unknown model type')
        return self._enhance_model(model=model, in_channels=in_nc, out_channels=out_nc, upscale=scale, name=name)



from typing import Mapping

from .arch import MoESR
from ...factory import KeyCondition, Architecture
from ...utilities.state_dict import get_seq_len


class MoESRArch(Architecture[MoESR]):
    def __init__(self):
        super().__init__(
            uid='MoESR',
            detect=KeyCondition.has_all(
                'in_to_dim.weight',
                'in_to_dim.bias',
                'blocks.0.blocks.0.gamma',
                'blocks.0.blocks.0.norm.weight',
                'blocks.0.blocks.0.norm.bias',
                'blocks.0.blocks.0.fc1.weight',
                'blocks.0.blocks.0.fc1.bias',
                'blocks.0.blocks.0.conv.dwconv_hw.weight',
                'blocks.0.blocks.0.conv.dwconv_hw.bias',
                'blocks.0.blocks.0.conv.dwconv_w.weight',
                'blocks.0.blocks.0.conv.dwconv_w.bias',
                'blocks.0.blocks.0.conv.dwconv_h.weight',
                'blocks.0.blocks.0.conv.dwconv_h.bias',
                'blocks.0.blocks.0.fc2.weight',
                'blocks.0.blocks.0.fc2.bias',
                'upscale.MetaUpsample',
            ),
        )

    def load(self, state: Mapping[str, object]):
        upsample = ['conv', 'pixelshuffledirect', 'pixelshuffle', 'nearest+conv', 'dysample']
        dim, in_ch = state['in_to_dim.weight'].shape[:2]
        n_blocks = get_seq_len(state, 'blocks')
        n_block = get_seq_len(state, 'blocks.0.blocks')
        expansion_factor_shape = state['blocks.0.blocks.0.fc1.weight'].shape
        expansion_factor = (expansion_factor_shape[0] / expansion_factor_shape[1]) / 2
        expansion_msg_shape = state['blocks.0.msg.gated.0.fc1.weight'].shape
        expansion_msg = (expansion_msg_shape[0] / expansion_msg_shape[1]) / 2
        _, index, scale, _, out_ch, upsample_dim, _ = state['upscale.MetaUpsample']
        upsampler = upsample[int(index)]

        model = MoESR(
            in_ch=in_ch,
            out_ch=int(out_ch),
            scale=int(scale),
            n_blocks=n_blocks,
            n_block=n_block,
            dim=dim,
            expansion_factor=expansion_factor,
            expansion_msg=expansion_msg,
            upsampler=upsampler,
            upsample_dim=int(upsample_dim),
        )

        return self._enhance_model(model=model, in_channels=in_ch, out_channels=int(out_ch), upscale=int(scale), name='MoESR')



from __future__ import annotations

from typing import Literal, Mapping


class KeyCondition:
    """
    A condition that checks if a state dict has the given keys.
    """

    def __init__(self, kind: Literal['all', 'any'], keys: tuple[str | KeyCondition, ...]):
        self._keys = keys
        self._kind: Literal['all', 'any'] = kind

    @staticmethod
    def has_all(*keys: str | KeyCondition) -> KeyCondition:
        return KeyCondition('all', keys)

    @staticmethod
    def has_any(*keys: str | KeyCondition) -> KeyCondition:
        return KeyCondition('any', keys)

    def __call__(self, state_dict: Mapping[str, object]) -> bool:
        def _detect(key: str | KeyCondition) -> bool:
            if isinstance(key, KeyCondition):
                return key(state_dict)
            return key in state_dict

        if self._kind == 'all':
            return all(_detect(key) for key in self._keys)
        else:
            return any(_detect(key) for key in self._keys)



import torch
import torch.nn.functional as F
from torch import nn
from torch.nn.init import trunc_normal_

from ...utilities import DySample


class Conv3XC(nn.Module):
    def __init__(self, c_in: int, c_out: int, gain: int = 1, s: int = 1, bias: bool = True):
        super(Conv3XC, self).__init__()
        self.weight_concat = None
        self.bias_concat = None
        self.update_params_flag = False
        self.stride = s

        self.sk = nn.Conv2d(
            in_channels=c_in,
            out_channels=c_out,
            kernel_size=1,
            padding=0,
            stride=s,
            bias=bias,
        )
        self.conv = nn.Sequential(
            nn.Conv2d(
                in_channels=c_in,
                out_channels=c_in * gain,
                kernel_size=1,
                padding=0,
                bias=bias,
            ),
            nn.Conv2d(
                in_channels=c_in * gain,
                out_channels=c_out * gain,
                kernel_size=3,
                stride=s,
                padding=0,
                bias=bias,
            ),
            nn.Conv2d(
                in_channels=c_out * gain,
                out_channels=c_out,
                kernel_size=1,
                padding=0,
                bias=bias,
            ),
        )
        self.eval_conv = nn.Conv2d(
            in_channels=c_in,
            out_channels=c_out,
            kernel_size=3,
            padding=1,
            stride=s,
            bias=bias,
        )
        if self.training:
            trunc_normal_(self.sk.weight, std=0.02)
            trunc_normal_(self.eval_conv.weight, std=0.02)

        if self.training is False:
            self.eval_conv.weight.requires_grad = False
            self.eval_conv.bias.requires_grad = False
            self.update_params()

    def update_params(self):
        w1 = self.conv[0].weight.data.clone().detach()
        b1 = self.conv[0].bias.data.clone().detach()
        w2 = self.conv[1].weight.data.clone().detach()
        b2 = self.conv[1].bias.data.clone().detach()
        w3 = self.conv[2].weight.data.clone().detach()
        b3 = self.conv[2].bias.data.clone().detach()

        w = F.conv2d(w1.flip(2, 3).permute(1, 0, 2, 3), w2, padding=2, stride=1).flip(2, 3).permute(1, 0, 2, 3)
        b = (w2 * b1.reshape(1, -1, 1, 1)).sum((1, 2, 3)) + b2

        self.weight_concat = F.conv2d(w.flip(2, 3).permute(1, 0, 2, 3), w3, padding=0, stride=1).flip(2, 3).permute(1, 0, 2, 3)
        self.bias_concat = (w3 * b.reshape(1, -1, 1, 1)).sum((1, 2, 3)) + b3

        sk_w = self.sk.weight.data.clone().detach()
        sk_b = self.sk.bias.data.clone().detach()
        target_kernel_size = 3

        H_pixels_to_pad = (target_kernel_size - 1) // 2
        W_pixels_to_pad = (target_kernel_size - 1) // 2
        sk_w = F.pad(sk_w, [H_pixels_to_pad, H_pixels_to_pad, W_pixels_to_pad, W_pixels_to_pad])

        self.weight_concat = self.weight_concat + sk_w
        self.bias_concat = self.bias_concat + sk_b

        self.eval_conv.weight.data = self.weight_concat
        self.eval_conv.bias.data = self.bias_concat

    def forward(self, x):
        if self.training:
            x_pad = F.pad(x, (1, 1, 1, 1), 'constant', 0)
            out = self.conv(x_pad) + self.sk(x)
        else:
            self.update_params()
            out = self.eval_conv(x)

        return out


class SPAB(nn.Module):
    def __init__(self, in_channels: int, end: bool = False):
        super(SPAB, self).__init__()

        self.in_channels = in_channels
        self.c1_r = Conv3XC(in_channels, in_channels, gain=2, s=1)
        self.c2_r = Conv3XC(in_channels, in_channels, gain=2, s=1)
        self.c3_r = Conv3XC(in_channels, in_channels, gain=2, s=1)
        self.act1 = nn.Mish(inplace=True)
        self.sigmoid = nn.Sigmoid()
        self.end = end

    def forward(self, x):
        out1 = self.c1_r(x)
        out1_act = self.act1(out1)

        out2 = self.c2_r(out1_act)
        out2_act = self.act1(out2)

        out3 = self.c3_r(out2_act)

        sim_att = self.sigmoid(out3) - 0.5
        out = (out3 + x) * sim_att
        if self.end:
            return out, out1
        return out


class SPABS(nn.Module):
    def __init__(self, feature_channels: int, n_blocks: int = 4, drop: float = 0.0):
        super(SPABS, self).__init__()
        self.block_1 = SPAB(feature_channels)

        self.block_n = nn.Sequential(*[SPAB(feature_channels) for _ in range(n_blocks)])
        self.block_end = SPAB(feature_channels, True)
        self.conv_2 = Conv3XC(feature_channels, feature_channels, gain=2, s=1)
        self.conv_cat = nn.Conv2d(feature_channels * 4, feature_channels, kernel_size=1, bias=True)
        self.dropout = nn.Dropout2d(drop)
        if self.training:
            trunc_normal_(self.conv_cat.weight, std=0.02)

    def forward(self, x):
        out_b1 = self.block_1(x)
        out_x = self.block_n(out_b1)
        out_end, out_x_2 = self.block_end(out_x)
        out_end = self.dropout(self.conv_2(out_end))
        return self.conv_cat(torch.cat([x, out_end, out_b1, out_x_2], 1))


class SpanPlus(nn.Module):
    """
    Swift Parameter-free Attention Network for Efficient Super-Resolution
    """

    def __init__(
        self,
        num_in_ch: int = 3,
        num_out_ch: int = 3,
        blocks: list[int] | int = [4],
        feature_channels: int = 48,
        upscale: int = 4,
        drop_rate: float = 0.0,
        upsampler: str = 'dys',  # "lp", "ps", "conv"- only 1x
    ):
        super(SpanPlus, self).__init__()

        in_channels = num_in_ch
        out_channels = num_out_ch if upsampler == 'dys' else num_in_ch
        if not isinstance(blocks, list):
            blocks = [int(blocks)]
        if not self.training:
            drop_rate = 0
        self.feats = nn.Sequential(
            *[Conv3XC(in_channels, feature_channels, gain=2, s=1)] + [SPABS(feature_channels, n_blocks, drop_rate) for n_blocks in blocks]
        )
        if upsampler == 'ps':
            self.upsampler = nn.Sequential(
                nn.Conv2d(feature_channels, out_channels * (upscale**2), 3, padding=1),
                nn.PixelShuffle(upscale),
            )
        elif upsampler == 'dys':
            self.upsampler = DySample(feature_channels, out_channels, upscale)
        elif upsampler == 'conv':
            if upscale != 1:
                msg = 'conv supports only 1x'
                raise ValueError(msg)

            self.upsampler = nn.Conv2d(feature_channels, out_channels, 3, padding=1)
        else:
            raise NotImplementedError(
                f'upsampler: {upsampler} not supported, choose one of these options: \
                ["ps", "dys", "conv"] conv supports only 1x'
            )

    def forward(self, x):
        out = self.feats(x)
        return self.upsampler(out)



import os
from typing import Mapping, Dict, Iterator, TypeVar

import torch
import pickle
from types import SimpleNamespace

from typing_extensions import override

from .factory import Architecture
from .utilities.state_dict import canonicalize_state_dict


class ArchitectureNotFound(Exception):
    pass


T = TypeVar('T', bound=torch.nn.Module, covariant=True)

safe_list = {
    ('collections', 'OrderedDict'),
    ('typing', 'OrderedDict'),
    ('torch._utils', '_rebuild_tensor_v2'),
    ('torch', 'BFloat16Storage'),
    ('torch', 'FloatStorage'),
    ('torch', 'HalfStorage'),
    ('torch', 'IntStorage'),
    ('torch', 'LongStorage'),
    ('torch', 'DoubleStorage'),
}


class RestrictedUnpickler(pickle.Unpickler):
    @override
    def find_class(self, module: str, name: str):
        # Only allow required classes to load state dict
        if (module, name) not in safe_list:
            raise pickle.UnpicklingError(f"Global '{module}.{name}' is forbidden")
        return super().find_class(module, name)


RestrictedUnpickle = SimpleNamespace(
    Unpickler=RestrictedUnpickler,
    __name__='pickle',
    load=lambda *args, **kwargs: RestrictedUnpickler(*args, **kwargs).load(),
)


class Registry:
    def __init__(self):
        self.store: Dict[str, Architecture] = {}

    def __contains__(self, id: str):
        return id in self.store

    def __iter__(self) -> Iterator[Architecture]:
        self._iter_keys = iter(self.store)
        return self

    def __next__(self) -> Architecture:
        if self._iter_keys is None:
            raise StopIteration
        try:
            key = next(self._iter_keys)
            return self.store[key]
        except StopIteration:
            self._iter_keys = None
            raise

    def add(self, cls: Architecture):
        self.store[cls.id] = cls

    def get(self, uid: str) -> Architecture:
        architecture = self.store[uid]
        if not architecture:
            raise ArchitectureNotFound
        return architecture

    def load_from_file(self, path: str) -> T:
        extension = os.path.splitext(path)[1].lower()
        if extension == '.pt':
            try:
                state_dict = torch.jit.load(path).state_dict()
            except RuntimeError:
                try:
                    pth_state_dict = torch.load(path, pickle_module=RestrictedUnpickle)
                except Exception:
                    pth_state_dict = None

                if pth_state_dict is None:
                    raise

                state_dict = pth_state_dict

        elif extension == '.pth' or extension == '.ckpt':
            state_dict = torch.load(path, pickle_module=RestrictedUnpickle)
        elif extension == '.safetensors':
            import safetensors.torch

            state_dict = safetensors.torch.load_file(path)
        else:
            raise ValueError(f'Unsupported model file extension {extension}. Please try a supported model type.')

        return self.load_from_state_dict(state_dict)

    def load_from_state_dict(self, state_dict: Mapping[str, object]) -> T:
        state_dict = canonicalize_state_dict(state_dict)

        for architecture in self.store.values():
            is_valid = architecture.detect(state_dict)
            if is_valid:
                model = architecture.load(state_dict)
                model.load_state_dict(state_dict)
                return model

        raise ArchitectureNotFound



from . import block
from . import state_dict
from .dysample import DySample
from .drop_path import DropPath
from .torch_internals import to_2tuple
from .padding import pad_to_multiple

__all__ = ['DySample', 'DropPath', 'to_2tuple', 'pad_to_multiple', 'block', 'state_dict']



import math
from typing import Literal

import torch
import torch.nn.functional as F
from einops import rearrange
from torch import Tensor, nn
import numpy as np

SampleMods = Literal['conv', 'pixelshuffledirect', 'pixelshuffle', 'nearest+conv', 'dysample', 'transpose+conv', 'lda', 'pa_up']


class DySample(nn.Module):
    """Adapted from 'Learning to Upsample by Learning to Sample':
    https://arxiv.org/abs/2308.15085
    https://github.com/tiny-smart/dysample
    """

    def __init__(
        self,
        in_channels: int = 64,
        out_ch: int = 3,
        scale: int = 2,
        groups: int = 4,
        end_convolution: bool = True,
        end_kernel=1,
    ) -> None:
        super().__init__()

        if in_channels <= groups or in_channels % groups != 0:
            msg = 'Incorrect in_channels and groups values.'
            raise ValueError(msg)

        out_channels = 2 * groups * scale**2
        self.scale = scale
        self.groups = groups
        self.end_convolution = end_convolution
        if end_convolution:
            self.end_conv = nn.Conv2d(in_channels, out_ch, end_kernel, 1, end_kernel // 2)
        self.offset = nn.Conv2d(in_channels, out_channels, 1)
        self.scope = nn.Conv2d(in_channels, out_channels, 1, bias=False)
        if self.training:
            nn.init.trunc_normal_(self.offset.weight, std=0.02)
            nn.init.constant_(self.scope.weight, val=0)

        self.register_buffer('init_pos', self._init_pos())

    def _init_pos(self) -> Tensor:
        h = torch.arange((-self.scale + 1) / 2, (self.scale - 1) / 2 + 1) / self.scale
        return torch.stack(torch.meshgrid([h, h], indexing='ij')).transpose(1, 2).repeat(1, self.groups, 1).reshape(1, -1, 1, 1)

    def forward(self, x: Tensor) -> Tensor:
        offset = self.offset(x) * self.scope(x).sigmoid() * 0.5 + self.init_pos
        B, _, H, W = offset.shape
        offset = offset.view(B, 2, -1, H, W)
        coords_h = torch.arange(H) + 0.5
        coords_w = torch.arange(W) + 0.5

        coords = (
            torch.stack(torch.meshgrid([coords_w, coords_h], indexing='ij'))
            .transpose(1, 2)
            .unsqueeze(1)
            .unsqueeze(0)
            .type(x.dtype)
            .to(x.device, non_blocking=True)
        )
        normalizer = torch.tensor([W, H], dtype=x.dtype, device=x.device, pin_memory=True).view(1, 2, 1, 1, 1)
        coords = 2 * (coords + offset) / normalizer - 1

        coords = (
            F.pixel_shuffle(coords.reshape(B, -1, H, W), self.scale)
            .view(B, 2, -1, self.scale * H, self.scale * W)
            .permute(0, 2, 3, 4, 1)
            .contiguous()
            .flatten(0, 1)
        )
        output = F.grid_sample(
            x.reshape(B * self.groups, -1, H, W),
            coords,
            mode='bilinear',
            align_corners=False,
            padding_mode='border',
        ).view(B, -1, self.scale * H, self.scale * W)

        if self.end_convolution:
            output = self.end_conv(output)

        return output


class LayerNorm(nn.Module):
    def __init__(self, dim: int = 64, eps: float = 1e-6) -> None:
        super().__init__()
        self.weight = nn.Parameter(torch.ones(dim))
        self.bias = nn.Parameter(torch.zeros(dim))
        self.eps = eps
        self.dim = (dim,)

    def forward(self, x):
        if x.is_contiguous(memory_format=torch.channels_last):
            return F.layer_norm(x.permute(0, 2, 3, 1), self.dim, self.weight, self.bias, self.eps).permute(0, 3, 1, 2)
        u = x.mean(1, keepdim=True)
        s = (x - u).pow(2).mean(1, keepdim=True)
        x = (x - u) / torch.sqrt(s + self.eps)
        return self.weight[:, None, None] * x + self.bias[:, None, None]


class LDA_AQU(nn.Module):
    def __init__(
        self,
        in_channels=48,
        reduction_factor=4,
        nh=1,
        scale_factor=2.0,
        k_e=3,
        k_u=3,
        n_groups=2,
        range_factor=11,
        rpb=True,
    ):
        super(LDA_AQU, self).__init__()
        self.k_u = k_u
        self.num_head = nh
        self.scale_factor = scale_factor
        self.n_groups = n_groups
        self.offset_range_factor = range_factor

        self.attn_dim = in_channels // (reduction_factor * self.num_head)
        self.scale = self.attn_dim**-0.5
        self.rpb = rpb
        self.hidden_dim = in_channels // reduction_factor
        self.proj_q = nn.Conv2d(in_channels, self.hidden_dim, kernel_size=1, stride=1, padding=0, bias=False)

        self.proj_k = nn.Conv2d(in_channels, self.hidden_dim, kernel_size=1, stride=1, padding=0, bias=False)

        self.group_channel = in_channels // (reduction_factor * self.n_groups)
        # print(self.group_channel)
        self.conv_offset = nn.Sequential(
            nn.Conv2d(
                self.group_channel,
                self.group_channel,
                3,
                1,
                1,
                groups=self.group_channel,
                bias=False,
            ),
            LayerNorm(self.group_channel),
            nn.SiLU(),
            nn.Conv2d(self.group_channel, 2 * k_u**2, k_e, 1, k_e // 2),
        )
        print(2 * k_u**2)
        self.layer_norm = LayerNorm(in_channels)

        self.pad = int((self.k_u - 1) / 2)
        base = np.arange(-self.pad, self.pad + 1).astype(np.float32)
        base_y = np.repeat(base, self.k_u)
        base_x = np.tile(base, self.k_u)
        base_offset = np.stack([base_y, base_x], axis=1).flatten()
        base_offset = torch.tensor(base_offset).view(1, -1, 1, 1)
        self.register_buffer('base_offset', base_offset, persistent=False)

        if self.rpb:
            self.relative_position_bias_table = nn.Parameter(torch.zeros(1, self.num_head, 1, self.k_u**2, self.hidden_dim // self.num_head))
            nn.init.trunc_normal_(self.relative_position_bias_table, std=0.02)

    def init_weights(self):
        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                nn.init.xavier_uniform(m)
            elif isinstance(m, nn.LayerNorm):
                nn.init.constant_(m.bias, 0)
                nn.init.constant_(m.weight, 1.0)
        nn.init.constant_(self.conv_offset[-1].weight, 0)
        nn.init.constant_(self.conv_offset[-1].bias, 0)

    def get_offset(self, offset, Hout, Wout):
        B, _, _, _ = offset.shape
        device = offset.device
        row_indices = torch.arange(Hout, device=device)
        col_indices = torch.arange(Wout, device=device)
        row_indices, col_indices = torch.meshgrid(row_indices, col_indices)
        index_tensor = torch.stack((row_indices, col_indices), dim=-1).view(1, Hout, Wout, 2)
        offset = rearrange(offset, 'b (kh kw d) h w -> b kh h kw w d', kh=self.k_u, kw=self.k_u)
        offset = offset + index_tensor.view(1, 1, Hout, 1, Wout, 2)
        offset = offset.contiguous().view(B, self.k_u * Hout, self.k_u * Wout, 2)

        offset[..., 0] = 2 * offset[..., 0] / (Hout - 1) - 1
        offset[..., 1] = 2 * offset[..., 1] / (Wout - 1) - 1
        offset = offset.flip(-1)
        return offset

    def extract_feats(self, x, offset, ks=3):
        out = nn.functional.grid_sample(x, offset, mode='bilinear', padding_mode='zeros', align_corners=True)
        out = rearrange(out, 'b c (ksh h) (ksw w) -> b (ksh ksw) c h w', ksh=ks, ksw=ks)
        return out

    def forward(self, x):
        B, C, H, W = x.shape
        out_H, out_W = int(H * self.scale_factor), int(W * self.scale_factor)
        v = x
        x = self.layer_norm(x)
        q = self.proj_q(x)
        k = self.proj_k(x)

        q = torch.nn.functional.interpolate(q, (out_H, out_W), mode='bilinear', align_corners=True)
        q_off = q.view(B * self.n_groups, -1, out_H, out_W)
        pred_offset = self.conv_offset(q_off)
        offset = pred_offset.tanh().mul(self.offset_range_factor) + self.base_offset.to(x.dtype)

        k = k.view(B * self.n_groups, self.hidden_dim // self.n_groups, H, W)
        v = v.view(B * self.n_groups, C // self.n_groups, H, W)
        offset = self.get_offset(offset, out_H, out_W)
        k = self.extract_feats(k, offset=offset)
        v = self.extract_feats(v, offset=offset)

        q = rearrange(q, 'b (nh c) h w -> b nh (h w) () c', nh=self.num_head)
        k = rearrange(k, '(b g) n c h w -> b (h w) n (g c)', g=self.n_groups)
        v = rearrange(v, '(b g) n c h w -> b (h w) n (g c)', g=self.n_groups)
        k = rearrange(k, 'b n1 n (nh c) -> b nh n1 n c', nh=self.num_head)
        v = rearrange(v, 'b n1 n (nh c) -> b nh n1 n c', nh=self.num_head)

        if self.rpb:
            k = k + self.relative_position_bias_table

        q = q * self.scale
        attn = q @ k.transpose(-1, -2)
        attn = attn.softmax(dim=-1)
        out = attn @ v

        out = rearrange(out, 'b nh (h w) t c -> b (nh c) (t h) w', h=out_H)
        return out


class PA(nn.Module):
    def __init__(self, dim):
        super(PA, self).__init__()
        self.conv = nn.Sequential(nn.Conv2d(dim, dim, 1), nn.Sigmoid())

    def forward(self, x):
        return x.mul(self.conv(x))


class UniUpsampleV3(nn.Sequential):
    def __init__(
        self,
        upsample: SampleMods = 'pa_up',
        scale: int = 2,
        in_dim: int = 48,
        out_dim: int = 3,
        mid_dim: int = 48,
        group: int = 4,  # Only DySample
        dysample_end_kernel=1,  # needed only for compatibility with version 2
    ) -> None:
        m = []

        if scale == 1 or upsample == 'conv':
            m.append(nn.Conv2d(in_dim, out_dim, 3, 1, 1))
        elif upsample == 'pixelshuffledirect':
            m.extend([nn.Conv2d(in_dim, out_dim * scale**2, 3, 1, 1), nn.PixelShuffle(scale)])
        elif upsample == 'pixelshuffle':
            m.extend([nn.Conv2d(in_dim, mid_dim, 3, 1, 1), nn.LeakyReLU(inplace=True)])
            if (scale & (scale - 1)) == 0:  # scale = 2^n
                for _ in range(int(math.log2(scale))):
                    m.extend([nn.Conv2d(mid_dim, 4 * mid_dim, 3, 1, 1), nn.PixelShuffle(2)])
            elif scale == 3:
                m.extend([nn.Conv2d(mid_dim, 9 * mid_dim, 3, 1, 1), nn.PixelShuffle(3)])
            else:
                raise ValueError(f'scale {scale} is not supported. Supported scales: 2^n and 3.')
            m.append(nn.Conv2d(mid_dim, out_dim, 3, 1, 1))
        elif upsample == 'nearest+conv':
            if (scale & (scale - 1)) == 0:
                for _ in range(int(math.log2(scale))):
                    m.extend(
                        (
                            nn.Conv2d(in_dim, in_dim, 3, 1, 1),
                            nn.Upsample(scale_factor=2),
                            nn.LeakyReLU(negative_slope=0.2, inplace=True),
                        )
                    )
                m.extend(
                    (
                        nn.Conv2d(in_dim, in_dim, 3, 1, 1),
                        nn.LeakyReLU(negative_slope=0.2, inplace=True),
                    )
                )
            elif scale == 3:
                m.extend(
                    (
                        nn.Conv2d(in_dim, in_dim, 3, 1, 1),
                        nn.Upsample(scale_factor=scale),
                        nn.LeakyReLU(negative_slope=0.2, inplace=True),
                        nn.Conv2d(in_dim, in_dim, 3, 1, 1),
                        nn.LeakyReLU(negative_slope=0.2, inplace=True),
                    )
                )
            else:
                raise ValueError(f'scale {scale} is not supported. Supported scales: 2^n and 3.')
            m.append(nn.Conv2d(in_dim, out_dim, 3, 1, 1))
        elif upsample == 'dysample':
            if mid_dim != in_dim:
                m.extend([nn.Conv2d(in_dim, mid_dim, 3, 1, 1), nn.LeakyReLU(inplace=True)])
            m.append(DySample(mid_dim, out_dim, scale, group, end_kernel=dysample_end_kernel))
            # m.append(nn.Conv2d(mid_dim, out_dim, dysample_end_kernel, 1, dysample_end_kernel//2)) # kernel 1 causes chromatic artifacts
        elif upsample == 'transpose+conv':
            if scale == 2:
                m.append(nn.ConvTranspose2d(in_dim, out_dim, 4, 2, 1))
            elif scale == 3:
                m.append(nn.ConvTranspose2d(in_dim, out_dim, 3, 3, 0))
            elif scale == 4:
                m.extend(
                    [
                        nn.ConvTranspose2d(in_dim, in_dim, 4, 2, 1),
                        nn.GELU(),
                        nn.ConvTranspose2d(in_dim, out_dim, 4, 2, 1),
                    ]
                )
            else:
                raise ValueError(f'scale {scale} is not supported. Supported scales: 2, 3, 4')
            m.append(nn.Conv2d(out_dim, out_dim, 3, 1, 1))
        elif upsample == 'lda':
            if mid_dim != in_dim:
                m.extend([nn.Conv2d(in_dim, mid_dim, 3, 1, 1), nn.LeakyReLU(inplace=True)])
            m.append(LDA_AQU(mid_dim, scale_factor=scale))
            m.append(nn.Conv2d(mid_dim, out_dim, 3, 1, 1))
        elif upsample == 'pa_up':
            if (scale & (scale - 1)) == 0:
                for _ in range(int(math.log2(scale))):
                    m.extend(
                        [
                            nn.Upsample(scale_factor=2),
                            nn.Conv2d(in_dim, mid_dim, 3, 1, 1),
                            PA(mid_dim),
                            nn.LeakyReLU(negative_slope=0.2, inplace=True),
                            nn.Conv2d(mid_dim, mid_dim, 3, 1, 1),
                            nn.LeakyReLU(negative_slope=0.2, inplace=True),
                        ]
                    )
                    in_dim = mid_dim
            elif scale == 3:
                m.extend(
                    [
                        nn.Upsample(scale_factor=3),
                        nn.Conv2d(in_dim, mid_dim, 3, 1, 1),
                        PA(mid_dim),
                        nn.LeakyReLU(negative_slope=0.2, inplace=True),
                        nn.Conv2d(mid_dim, mid_dim, 3, 1, 1),
                        nn.LeakyReLU(negative_slope=0.2, inplace=True),
                    ]
                )
            else:
                raise ValueError(f'scale {scale} is not supported. Supported scales: 2^n and 3.')
            m.append(nn.Conv2d(mid_dim, out_dim, 3, 1, 1))
        else:
            raise ValueError(f'An invalid Upsample was selected. Please choose one of {SampleMods}')
        super().__init__(*m)

        self.register_buffer(
            'MetaUpsample',
            torch.tensor(
                [
                    3,  # Block version, if you change something, please number from the end so that you can distinguish between authorized changes and third parties
                    list(SampleMods.__args__).index(upsample),  # UpSample method index
                    scale,
                    in_dim,
                    out_dim,
                    mid_dim,
                    group,
                ],
                dtype=torch.uint8,
            ),
        )


class Conv3XC(nn.Module):
    def __init__(self, c_in, c_out, gain1=1, s=1, bias=True) -> None:
        super().__init__()
        self.bias = bias
        self.weight_concat = None
        self.bias_concat = None
        self.update_params_flag = False
        self.stride = s
        gain = gain1

        self.sk = nn.Conv2d(
            in_channels=c_in,
            out_channels=c_out,
            kernel_size=1,
            padding=0,
            stride=s,
            bias=bias,
        )
        self.conv = nn.Sequential(
            nn.Conv2d(
                in_channels=c_in,
                out_channels=c_in * gain,
                kernel_size=1,
                padding=0,
                bias=bias,
            ),
            nn.Conv2d(
                in_channels=c_in * gain,
                out_channels=c_out * gain,
                kernel_size=3,
                stride=s,
                padding=0,
                bias=bias,
            ),
            nn.Conv2d(
                in_channels=c_out * gain,
                out_channels=c_out,
                kernel_size=1,
                padding=0,
                bias=bias,
            ),
        )

        self.eval_conv = nn.Conv2d(
            in_channels=c_in,
            out_channels=c_out,
            kernel_size=3,
            padding=1,
            stride=s,
            bias=bias,
        )
        nn.init.trunc_normal_(self.sk.weight, std=0.02)
        if self.training is False:
            self.eval_conv.weight.requires_grad = False
            self.eval_conv.bias.requires_grad = False
            self.update_params()

    def update_params(self) -> None:
        w1 = self.conv[0].weight.data.clone().detach()
        w2 = self.conv[1].weight.data.clone().detach()
        w3 = self.conv[2].weight.data.clone().detach()
        w = F.conv2d(w1.flip(2, 3).permute(1, 0, 2, 3), w2, padding=2, stride=1).flip(2, 3).permute(1, 0, 2, 3)

        self.weight_concat = F.conv2d(w.flip(2, 3).permute(1, 0, 2, 3), w3, padding=0, stride=1).flip(2, 3).permute(1, 0, 2, 3)

        sk_w = self.sk.weight.data.clone().detach()

        if self.bias:
            b1 = self.conv[0].bias.data.clone().detach()
            b2 = self.conv[1].bias.data.clone().detach()
            b3 = self.conv[2].bias.data.clone().detach()
            b = (w2 * b1.reshape(1, -1, 1, 1)).sum((1, 2, 3)) + b2
            self.bias_concat = (w3 * b.reshape(1, -1, 1, 1)).sum((1, 2, 3)) + b3
            sk_b = self.sk.bias.data.clone().detach()

        target_kernel_size = 3

        H_pixels_to_pad = (target_kernel_size - 1) // 2
        W_pixels_to_pad = (target_kernel_size - 1) // 2
        sk_w = F.pad(sk_w, [H_pixels_to_pad, H_pixels_to_pad, W_pixels_to_pad, W_pixels_to_pad])
        self.weight_concat = self.weight_concat + sk_w
        self.eval_conv.weight.data = self.weight_concat
        if self.bias:
            self.bias_concat = self.bias_concat + sk_b
            self.eval_conv.bias.data = self.bias_concat

    def train(self, mode: bool = True):
        super().train(mode)
        if not mode:
            self.update_params()
        return self

    def forward(self, x):
        if self.training:
            pad = 1
            x_pad = F.pad(x, (pad, pad, pad, pad), 'constant', 0)
            out = self.conv(x_pad) + self.sk(x)
        else:
            out = self.eval_conv(x)
        return out


class SPAB(nn.Module):
    def __init__(self, in_channels, mid_dim=None, out_dim=None, bias=False, end=False) -> None:
        super().__init__()
        mid_dim = mid_dim if mid_dim else in_channels
        out_dim = out_dim if out_dim else in_channels
        self.in_channels = in_channels
        self.c1_r = Conv3XC(in_channels, mid_dim, gain1=2, s=1, bias=bias)
        self.c2_r = Conv3XC(mid_dim, mid_dim, gain1=2, s=1, bias=bias)
        self.c3_r = Conv3XC(mid_dim, out_dim, gain1=2, s=1, bias=bias)
        self.act1 = torch.nn.SiLU(inplace=True)
        self.end = end

    def forward(self, x):
        out1 = self.c1_r(x)
        out1_act = self.act1(out1)
        out2 = self.c2_r(out1_act)
        out2_act = self.act1(out2)
        out3 = self.c3_r(out2_act)
        sim_att = torch.sigmoid(out3) - 0.5
        out = (out3 + x) * sim_att
        if self.end:
            return out, out1
        return out


class RMSNorm(nn.Module):
    def __init__(self, dim: int, eps: float = 1e-6) -> None:
        super().__init__()
        self.eps = eps
        self.scale = nn.Parameter(torch.ones(dim))
        self.offset = nn.Parameter(torch.zeros(dim))

    def forward(self, x):
        norm_x = x.norm(2, dim=1, keepdim=True)
        d_x = x.size(1)
        rms_x = norm_x * (d_x ** (-1.0 / 2))
        x_normed = x / (rms_x + self.eps)
        return self.scale[..., None, None] * x_normed + self.offset[..., None, None]


class InceptionDWConv2d(nn.Module):
    """Inception depthweise convolution"""

    def __init__(self, in_channels, square_kernel_size=3, band_kernel_size=11, branch_ratio=0.125) -> None:
        super().__init__()

        gc = int(in_channels * branch_ratio)  # channel numbers of a convolution branch
        self.dwconv_hw = nn.Conv2d(gc, gc, square_kernel_size, padding=square_kernel_size // 2, groups=gc)
        self.dwconv_w = nn.Conv2d(
            gc,
            gc,
            kernel_size=(1, band_kernel_size),
            padding=(0, band_kernel_size // 2),
            groups=gc,
        )
        self.dwconv_h = nn.Conv2d(
            gc,
            gc,
            kernel_size=(band_kernel_size, 1),
            padding=(band_kernel_size // 2, 0),
            groups=gc,
        )
        self.split_indexes = (in_channels - 3 * gc, gc, gc, gc)

    def forward(self, x):
        x_id, x_hw, x_w, x_h = torch.split(x, self.split_indexes, dim=1)
        return torch.cat(
            (x_id, self.dwconv_hw(x_hw), self.dwconv_w(x_w), self.dwconv_h(x_h)),
            dim=1,
        )


class Attention(nn.Module):
    def __init__(self, dim, num_heads) -> None:
        super().__init__()
        self.num_heads = num_heads
        self.head_dim = dim // num_heads
        assert dim % num_heads == 0, 'dim must be divisible by num_heads'
        self.temperature = nn.Parameter(torch.ones(num_heads, 1, 1))

        self.qkv = nn.Conv2d(dim, dim * 3, kernel_size=1, bias=False)
        self.qkv_dwconv = nn.Conv2d(dim * 3, dim * 3, 3, 1, 1, 1, dim * 3)
        self.project_out = nn.Conv2d(dim, dim, kernel_size=1, bias=False)

    def forward(self, x):
        b, c, h, w = x.shape

        qkv = self.qkv_dwconv(self.qkv(x))
        q, k, v = torch.chunk(qkv, 3, dim=1)

        # Преобразуем в (b, num_heads, head_dim, h*w)
        q = q.view(b, self.num_heads, self.head_dim, h * w)
        k = k.view(b, self.num_heads, self.head_dim, h * w)
        v = v.view(b, self.num_heads, self.head_dim, h * w)

        q = F.normalize(q, dim=3)
        k = F.normalize(k, dim=3)

        attn = torch.matmul(q, k.transpose(2, 3)) * self.temperature  # (b, num_heads, hw, hw)
        attn = attn.softmax(dim=3)

        out = torch.matmul(attn, v)  # (b, num_heads, head_dim, hw)

        # Обратно в (b, c, h, w)
        out = out.view(b, c, h, w)

        out = self.project_out(out)
        return out


class GatedCNNBlock(nn.Module):
    r"""
    modernized mambaout main unit
    https://github.com/yuweihao/MambaOut/blob/main/models/mambaout.py#L119
    """

    def __init__(
        self,
        dim: int = 64,
        expansion_ratio: float = 1.5,
        conv_ratio: float = 1,
        att=False,
    ) -> None:
        super().__init__()
        self.norm = RMSNorm(dim)
        hidden = int(expansion_ratio * dim)
        self.fc1 = nn.Conv2d(dim, hidden * 2, 1, 1)

        self.act = nn.Mish()
        conv_channels = int(conv_ratio * dim)
        self.split_indices = [hidden, hidden - conv_channels, conv_channels]
        self.token_mix = Attention(conv_channels, 16) if att else InceptionDWConv2d(dim)
        self.fc2 = nn.Conv2d(hidden, dim, 1, 1, 0)
        self.apply(self._init_weights)

    @staticmethod
    def _init_weights(m) -> None:
        if isinstance(m, nn.Conv2d | nn.Linear):
            nn.init.trunc_normal_(m.weight, std=0.02)
            if m.bias is not None:
                nn.init.constant_(m.bias, 0)

    def forward(self, x):
        x = self.norm(x)
        g, i, c = torch.split(self.fc1(x), self.split_indices, dim=1)
        c = self.token_mix(c)
        x = self.act(g) * torch.cat((i, c), dim=1)
        x = self.act(self.fc2(x))
        return x


class SimpleGate(nn.Module):
    @staticmethod
    def forward(x):
        x1, x2 = x.chunk(2, dim=1)
        return x1 * x2


class MetaGated(nn.Module):
    def __init__(self, dim) -> None:
        super().__init__()
        hidden = dim * 2
        self.local = nn.Sequential(
            RMSNorm(dim),
            nn.Conv2d(dim, hidden, 1),
            nn.Conv2d(hidden, hidden, 3, 1, 1, groups=dim),
            SimpleGate(),
        )
        self.sca = nn.Sequential(
            nn.AdaptiveAvgPool2d(1),
            nn.Conv2d(in_channels=dim, out_channels=dim, kernel_size=1),
        )
        self.glob = GatedCNNBlock(dim)
        self.gamma0 = nn.Parameter(torch.ones([1, dim, 1, 1]), requires_grad=True)
        self.gamma1 = nn.Parameter(torch.ones([1, dim, 1, 1]), requires_grad=True)
        self.apply(self._init_weights)

    @staticmethod
    def _init_weights(m) -> None:
        if isinstance(m, nn.Conv2d | nn.Linear):
            nn.init.trunc_normal_(m.weight, std=0.02)
            if m.bias is not None:
                nn.init.constant_(m.bias, 0)

    def forward(self, x):
        short = x
        x = self.local(x)
        x = x * self.sca(x)
        x = x * self.gamma0 + short
        del short
        x = self.glob(x) * self.gamma1 + x
        return x


class Down(nn.Sequential):
    def __init__(self, dim) -> None:
        super().__init__(nn.Conv2d(dim, dim // 2, 3, 1, 1, bias=False), nn.PixelUnshuffle(2))


class Upsample(nn.Sequential):
    def __init__(self, dim) -> None:
        super().__init__(nn.Conv2d(dim, dim * 2, 3, 1, 1, bias=False), nn.PixelShuffle(2))


class Block(nn.Module):
    def __init__(self, dim, num_gated, down=True) -> None:
        super().__init__()

        if down:
            self.gated = nn.Sequential(*[MetaGated(dim) for _ in range(num_gated)])
            self.scale = Down(dim)

        else:
            self.scale = Upsample(dim)
            self.gated = nn.Sequential(*[MetaGated(dim // 2) for _ in range(num_gated)])
            self.shor = nn.Conv2d(int(dim), dim // 2, 1, 1, 0)
        self.down = down

    def forward(self, x, short=None):
        if self.down:
            x = self.gated(x)
            return self.scale(x), x
        else:
            x = torch.cat([self.scale(x), short], dim=1)
            x = self.shor(x)
            return self.gated(x)


class GateRV3(nn.Module):
    def __init__(
        self,
        in_ch=3,
        dim=32,
        enc_blocks=(2, 2, 4, 8),
        dec_blocks=(2, 2, 2, 2),
        num_latent=12,
        scale=1,
        upsample: SampleMods = 'pixelshuffle',
        upsample_mid_dim=32,
        end_gamma_init=1,
        attention=False,
        span_blocks=4,
        end_kernel=3,
        **kwargs,
    ) -> None:
        super().__init__()

        self.scale = scale
        self.in_to_dim = nn.Conv2d(in_ch, dim, 3, 1, 1)
        self.gater_encode = nn.ModuleList([Block(dim * (2**i), enc_blocks[i]) for i in range(len(enc_blocks))])
        self.span_block0 = SPAB(dim, end=False)
        self.span_n_b = nn.Sequential(*[SPAB(dim, end=False) for _ in range(span_blocks)])
        self.span_end = SPAB(dim, end=True)
        self.sisr_end_conv = Conv3XC(dim, dim, bias=True)
        self.sisr_cat_conv = nn.Conv2d(dim * 4, dim, 1)
        nn.init.trunc_normal_(self.sisr_cat_conv.weight, std=0.02)
        self.latent = nn.Sequential(
            *[
                GatedCNNBlock(
                    dim * (2 ** len(enc_blocks)),
                    expansion_ratio=1.5,
                    conv_ratio=1.00,
                    att=attention,
                )
                for _ in range(num_latent)
            ]
        )
        self.decode = nn.ModuleList(
            [
                Block(
                    dim * (2 ** (len(dec_blocks) - i)),
                    dec_blocks[i],
                    False,
                )
                for i in range(len(dec_blocks))
            ]
        )
        self.pad = 2 ** (len(enc_blocks))
        # self.dim_to_in = nn.Conv2d(dim*2, in_ch, 3, 1, 1)
        self.gamma = nn.Parameter(torch.ones(1, in_ch, 1, 1) * end_gamma_init)
        self.gamma.register_hook(lambda grad: grad * 100)
        if scale != 1:
            self.short_to_dim = nn.Upsample(scale_factor=scale)  # ConvBlock(in_ch, dim)
            self.dim_to_in = UniUpsampleV3(upsample, scale, dim, in_ch, upsample_mid_dim, dysample_end_kernel=end_kernel)
            # self.upsample =
        else:
            self.dim_to_in = nn.Conv2d(dim, in_ch, 3, 1, 1)
            self.short_to_dim = nn.Identity()

    def load_state_dict(self, state_dict, *args, **kwargs):
        if 'dim_to_in.MetaUpsample' in state_dict:
            state_dict['dim_to_in.MetaUpsample'] = self.dim_to_in.MetaUpsample
        if 'gamma' not in state_dict:
            state_dict['gamma'] = self.gamma
        return super().load_state_dict(state_dict, *args, **kwargs)

    def check_img_size(self, x, resolution: tuple[int, int]):
        scaled_size = self.pad
        mod_pad_h = (scaled_size - resolution[0] % scaled_size) % scaled_size
        mod_pad_w = (scaled_size - resolution[1] % scaled_size) % scaled_size
        return F.pad(x, (0, mod_pad_w, 0, mod_pad_h), 'reflect')

    def forward(self, inp):
        B, C, H, W = inp.shape
        inp = self.check_img_size(inp, (H, W))
        x = self.in_to_dim(inp)
        sisr = self.span_block0(x)
        sisr_short = sisr
        sisr = self.span_n_b(sisr)
        sisr, sisr_out = self.span_end(sisr)
        sisr = self.sisr_end_conv(sisr)
        sisr = self.sisr_cat_conv(torch.cat([x, sisr, sisr_short, sisr_out], dim=1))
        del sisr_short, sisr_out
        shorts = []
        for block in self.gater_encode:
            x, short = block(x)
            shorts.append(short)

        x = self.latent(x)
        len_block = len(self.decode)
        shorts.reverse()
        for index in range(len_block):
            x = self.decode[index](x, shorts[index])

        x = self.dim_to_in(x + sisr) + self.gamma * self.short_to_dim(inp)
        return x[:, :, : H * self.scale, : W * self.scale]



import math
from typing import Mapping

from .arch import RGT
from ...factory import Architecture, KeyCondition
from ...utilities.state_dict import get_seq_len, get_pixelshuffle_params


def _get_split_size(state_dict: Mapping[str, object]) -> tuple[int, int]:
    # split_size can only be uniquely determined if split_size[0] == split_size[1]
    # From now on, we call ssw=split_size[0] and ssh=split_size[1]
    # What we know:
    #   a = ssw * ssh
    #   b = (ssw * 2 - 1) * (ssh * 2 - 1)
    # Since this is not enough to uniquely determine ssw and ssh, we assume:
    #   ssw <= ssh
    a = state_dict['layers.0.blocks.0.attn.attns.0.relative_position_index'].shape[0]
    b = state_dict['layers.0.blocks.0.attn.attns.0.rpe_biases'].shape[0]

    def is_solution(ssw: int, ssh: int) -> bool:
        return ssw * ssh == a and (2 * ssw - 1) * (2 * ssh - 1) == b

    # this simplifies a lot if ssw == ssh
    square_size = math.isqrt(a)
    if is_solution(square_size, square_size):
        return square_size, square_size

    # otherwise, we'll just check powers of 2
    for i in range(1, 10):
        ssw = 2**i
        for j in range(i + 1, 10):
            ssh = 2**j
            if is_solution(ssw, ssh):
                return ssw, ssh

    raise ValueError(f'No valid split_size found for {a=} and {b=}')


class RGTArch(Architecture[RGT]):
    def __init__(self):
        super().__init__(
            uid='RGT',
            detect=KeyCondition.has_all(
                'conv_first.weight',
                'before_RG.1.weight',
                'layers.0.blocks.0.gamma',
                'layers.0.blocks.0.norm1.weight',
                'layers.0.blocks.0.attn.qkv.weight',
                'layers.0.blocks.0.attn.proj.weight',
                'layers.0.blocks.0.attn.attns.0.rpe_biases',
                'layers.0.blocks.0.attn.attns.0.relative_position_index',
                'layers.0.blocks.0.attn.attns.0.pos.pos_proj.weight',
                'layers.0.blocks.0.mlp.fc1.weight',
                'layers.0.blocks.0.mlp.fc2.weight',
                'layers.0.blocks.0.norm2.weight',
                'norm.weight',
                KeyCondition.has_any(
                    # 1conv
                    'conv_after_body.weight',
                    # 3conv
                    'conv_after_body.0.weight',
                ),
                'conv_before_upsample.0.weight',
                'conv_last.weight',
            ),
        )

    def load(self, state_dict: Mapping[str, object]):
        img_size = 64  # unused
        qk_scale = None  # cannot be deduced from state_dict
        drop_rate = 0.0  # cannot be deduced from state_dict
        attn_drop_rate = 0.0  # cannot be deduced from state_dict
        drop_path_rate = 0.1  # cannot be deduced from state_dict
        img_range = 1.0  # cannot be deduced from state_dict
        c_ratio = 0.5

        in_chans = state_dict['conv_first.weight'].shape[1]
        embed_dim = state_dict['conv_first.weight'].shape[0]

        num_layers = get_seq_len(state_dict, 'layers')
        depth = [0] * num_layers
        num_heads = [2] * num_layers
        for i in range(num_layers):
            depth[i] = get_seq_len(state_dict, f'layers.{i}.blocks')
            heads_half = state_dict[f'layers.{i}.blocks.0.attn.attns.0.pos.pos3.2.weight'].shape[0]
            if embed_dim % (heads_half * 2) == 0:
                num_heads[i] = heads_half * 2
            else:
                num_heads[i] = heads_half * 2 + 1

        qkv_bias = 'layers.0.blocks.0.attn.qkv.bias' in state_dict

        mlp_ratio = state_dict['layers.0.blocks.0.mlp.fc1.weight'].shape[0] / state_dict['layers.0.blocks.0.mlp.fc1.weight'].shape[1]

        if 'conv_after_body.weight' in state_dict:
            resi_connection = '1conv'
        else:
            resi_connection = '3conv'

        # c_ratio is only defined if at least one depth is >= 2
        for i, d in enumerate(depth):
            if d >= 2:
                c_ratio = state_dict[f'layers.{i}.blocks.1.attn.conv.weight'].shape[0] / state_dict[f'layers.{i}.blocks.1.attn.conv.weight'].shape[1]
                break

        upscale, _ = get_pixelshuffle_params(state_dict, 'upsample')

        split_size = _get_split_size(state_dict)

        model = RGT(
            img_size=img_size,
            in_chans=in_chans,
            embed_dim=embed_dim,
            depth=depth,
            num_heads=num_heads,
            mlp_ratio=mlp_ratio,
            qkv_bias=qkv_bias,
            qk_scale=qk_scale,
            drop_rate=drop_rate,
            attn_drop_rate=attn_drop_rate,
            drop_path_rate=drop_path_rate,
            upscale=upscale,
            img_range=img_range,
            resi_connection=resi_connection,
            split_size=split_size,
            c_ratio=c_ratio,
        )
        return self._enhance_model(model=model, in_channels=in_chans, out_channels=in_chans, upscale=upscale, name='RGT')



from dataclasses import dataclass

import torch
from abc import ABC, abstractmethod
from typing import Generic, TypeVar, Mapping

from .key_condition import KeyCondition

T = TypeVar('T', bound=torch.nn.Module, covariant=True)


@dataclass
class ModelMetadata:
    """Mixin for adding SR model specific attributes and methods"""

    in_channels: int
    out_channels: int
    upscale: int
    name: str


class Architecture(ABC, Generic[T]):
    def __init__(self, uid: str, detect: KeyCondition):
        self.id = uid
        self._detect = detect

    def detect(self, state_dict: Mapping[str, object]) -> bool:
        return self._detect(state_dict)

    @abstractmethod
    def load(self, state_dict: Mapping[str, object]) -> T:
        raise NotImplementedError

    def _enhance_model(self, model: T, in_channels: int, out_channels: int, upscale: int, name) -> T:
        model.parameters_info = ModelMetadata(name=name, in_channels=in_channels, out_channels=out_channels, upscale=upscale)
        return model



from typing import Mapping

from .archs import internal_registry


def add(arch):
    """Register a new architecture."""
    return internal_registry.add(arch)


def get(id: str):
    """Get architecture by ID."""
    return internal_registry.get(id)


def load_from_file(path: str):
    """Detect and load architecture from state dict."""
    return internal_registry.load_from_file(path)


def load_from_state_dict(state_dict: Mapping[str, object]):
    """Detect and load architecture from state dict."""
    return internal_registry.load_from_state_dict(state_dict)


__all__ = ['add', 'get', 'load_from_file', 'load_from_state_dict']
